
//***************************************************
// Generalised Bitmap Module C++ adapter for GBM plugins
//
// Requires: GBM.DLL version 1.35 or higher (with multipage support)
//
// Copyright (C) 2006-2008 Heiko Nitzsche
//
//   This software is provided 'as-is', without any express or implied
//   warranty.  In no event will the author be held liable for any damages
//   arising from the use of this software.
//
//   Permission is granted to anyone to use this software for any purpose,
//   including commercial applications, and to alter it and redistribute it
//   freely, subject to the following restrictions:
//
//   1. The origin of this software must not be misrepresented; you must not
//      claim that you wrote the original software. If you use this software
//      in a product, an acknowledgment in the product documentation would be
//      appreciated but is not required.
//   2. Altered source versions must be plainly marked as such, and must not be
//      misrepresented as being the original software.
//   3. This notice may not be removed or altered from any source distribution.
//
//***************************************************

#ifdef __IBMCPP__
  #pragma strings( readonly )
#endif

#include "os2defs.h"

#include <assert.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <process.h>

#include "GbmDocument.hpp"

// ---------------------------------------------------------

// Should be multiple of 2
//#define DEFAULT_MAX_RENDER_PAGES  2;

// ---------------------------------------------------------

GbmDocument::BackgroundRenderQueue::BackgroundRenderQueue(
  GbmDocument * pGbmDocument)
  : mThreadId(-1)
{
  mThreadArgs.hmtx         = NULLHANDLE;
  mThreadArgs.hev          = NULLHANDLE;
  mThreadArgs.finish       = FALSE;
  mThreadArgs.pGbmDocument = pGbmDocument;

  mThreadArgs.pages    = NULL;
  mThreadArgs.numPages = 0;

#ifdef DEFAULT_MAX_RENDER_PAGES
  if (DEFAULT_MAX_RENDER_PAGES > 0)
  {
    mThreadArgs.pages = new int[DEFAULT_MAX_RENDER_PAGES];
    assert(mThreadArgs.pages != NULL);
    mThreadArgs.numPages = DEFAULT_MAX_RENDER_PAGES;
  }
#endif

  clearRendererList();

  DosCreateMutexSem(NULL, &(mThreadArgs.hmtx), 0L, FALSE);
  assert(mThreadArgs.hmtx != NULLHANDLE);

  DosCreateEventSem(NULL, &(mThreadArgs.hev), 0L, TRUE);
  assert(mThreadArgs.hev != NULLHANDLE);
}

// ------------------

GbmDocument::BackgroundRenderQueue::~BackgroundRenderQueue()
{
  // wait for render thread
  stop();

  DosCloseEventSem(mThreadArgs.hev);
  DosCloseMutexSem(mThreadArgs.hmtx);
  mThreadArgs.hev  = NULLHANDLE;
  mThreadArgs.hmtx = NULLHANDLE;
  mThreadArgs.pGbmDocument = NULL;

  delete [] mThreadArgs.pages;
  mThreadArgs.pages    = NULL;
  mThreadArgs.numPages = 0;
}

// ------------------

BOOL GbmDocument::BackgroundRenderQueue::setNumberOfBackgroundRenderPages(const int numCachePages)
{
  if (numCachePages < 1)
  {
      return FALSE;
  }
  int * pNewPages = new int[numCachePages];
  if (pNewPages == NULL)
  {
      return FALSE;
  }

  lock();

  delete [] mThreadArgs.pages;
  mThreadArgs.pages    = pNewPages;
  mThreadArgs.numPages = numCachePages;

  clearRendererList();

  unlock();

  return TRUE;
}

// ------------------

int GbmDocument::BackgroundRenderQueue::getNumberOfBackgroundRenderPages() const
{
    return mThreadArgs.numPages;
}

// ------------------

void GbmDocument::BackgroundRenderQueue::lock()
{
  // wait for render thread
  DosRequestMutexSem(mThreadArgs.hmtx, SEM_INDEFINITE_WAIT);
}

void GbmDocument::BackgroundRenderQueue::unlock()
{
  DosReleaseMutexSem(mThreadArgs.hmtx);
}

// ------------------

void GbmDocument::BackgroundRenderQueue::clearRendererList()
{
  for (int i = 0; i < mThreadArgs.numPages; i++)
  {
    mThreadArgs.pages[i] = -1;
  }
  mThreadArgs.scaleX = 0.0;
  mThreadArgs.scaleY = 0.0;
}

// ------------------

#ifdef __IBMCPP__
  static void _Optlink BackgroundRenderQueue_threadFunc(void *arg)
#else
  static void BackgroundRenderQueue_threadFunc(void *arg)
#endif
{
  GbmDocument::BackgroundRenderQueue::THREAD_ARG * pThreadArgs =
    (GbmDocument::BackgroundRenderQueue::THREAD_ARG *) arg;

  ULONG postCount;

  while ((!pThreadArgs->finish) && (DosWaitEventSem(pThreadArgs->hev, SEM_INDEFINITE_WAIT) == 0))
  {
    for (int i = 0; i < pThreadArgs->numPages; i++)
    {
      if (pThreadArgs->finish)
      {
        break;
      }
      DosRequestMutexSem(pThreadArgs->hmtx, SEM_INDEFINITE_WAIT);
      if (pThreadArgs->pages[i] != -1)
      {
        pThreadArgs->pGbmDocument->prepareRenderCache(pThreadArgs->pages[i],
                                                      pThreadArgs->scaleX,
                                                      pThreadArgs->scaleY);
        pThreadArgs->pages[i] = -1;
      }
      DosReleaseMutexSem(pThreadArgs->hmtx);
    }

    DosResetEventSem(pThreadArgs->hev, &postCount);
  }
  _endthread();
}

// ------------------

void GbmDocument::BackgroundRenderQueue::start()
{
  if (mThreadId == -1)
  {
    mThreadId = _beginthread(BackgroundRenderQueue_threadFunc,
                             NULL, 8192, &mThreadArgs);
//    DosSetPriority(PRTYS_THREAD, PRTYC_IDLETIME, PRTYD_MAXIMUM, mThreadId);
  }
}

// ------------------

void GbmDocument::BackgroundRenderQueue::stop()
{
  if (mThreadId != -1)
  {
    // wait for render thread
    mThreadArgs.finish = TRUE;
    DosPostEventSem(mThreadArgs.hev);
    DosWaitThread((PTID)&mThreadId, DCWW_WAIT);
    mThreadId = -1;

    clearRendererList();
  }
}

// ------------------

void GbmDocument::BackgroundRenderQueue::clear()
{
    lock();
    clearRendererList();
    unlock();
}

// ------------------

BOOL GbmDocument::BackgroundRenderQueue::schedulePages(
    int * pPages, int numPages, double scaleX, double scaleY)
{
    if (mThreadId == -1)
    {
        return FALSE;
    }

    // wait for render thread
    lock();

    // more renderer to schedule than queue size
    if (numPages > mThreadArgs.numPages)
    {
      unlock();
      return FALSE;
    }

    // replace the existing renderers
    clearRendererList();

    if (numPages > 0)
    {
      for (int i = 0; i < numPages; i++)
      {
        mThreadArgs.pages[i] = pPages[i];
      }
      mThreadArgs.scaleX = scaleX;
      mThreadArgs.scaleY = scaleY;
    }

    unlock();

    // wake up thread to continue rendering
    DosPostEventSem(mThreadArgs.hev);

    return TRUE;
}

// ---------------------------------------------------------
// ---------------------------------------------------------

static BOOL isEqualLowerCase(const char *s1, const char *s2, int n)
{
  for ( ; n--; s1++, s2++ )
  {
    if ( tolower(*s1) != tolower(*s2) )
    {
      return FALSE;
    }
  }
  return TRUE;
}

// ---------------------------------------------------------

static const char *findWord(const char *str, const char *substr)
{
         char * buf, *s;
         char   strbuf[51];
         int    len            = strlen(substr);
   const int    strlength      = strlen(str);
   const BOOL   needsDynBuffer = (strlength > (sizeof(strbuf)-1));

   /* only use dynamic buffer if needed to speedup a bit */
   if (needsDynBuffer)
   {
     buf = (char *) malloc(strlength + 1);
     if (buf == NULL)
     {
       return NULL;
     }
   }
   else
   {
     buf = strbuf;
   }

   for ( s  = strtok(strcpy(buf, str), " \t,");
         s != NULL;
         s  = strtok(NULL, " \t,") )
   {
     if ( isEqualLowerCase(s, substr, len) && s[len] == '\0' )
     {
       int inx = s - buf;
       if (needsDynBuffer)
       {
         free(buf);
       }
       return str + inx;
     }
   }
   if (needsDynBuffer)
   {
     free(buf);
   }
   return NULL;
}

// ---------------------------------------------------------

// Lookup GBM file type. If extension is specified, use this one
// and ignore the filename.

static int lookupGbmFileType(const GbmAccessor & gbmAccessor, const char * filename, const char * extension)
{
    if (extension != NULL)
    {
        int numFileTypes = -1;

        if (gbmAccessor.Gbm_query_n_filetypes(&numFileTypes) == GBM_ERR_OK)
        {
          int   ft;
          GBMFT gbmft;

          for (ft = 0; ft < numFileTypes; ft++)
          {
            if (gbmAccessor.Gbm_query_filetype(ft, &gbmft) == GBM_ERR_OK)
            {
              if (findWord(gbmft.extensions, extension) != NULL)
              {
                return ft;
              }
            }
          }
        }
    }
    else
    {
        int ft;
        if (gbmAccessor.Gbm_guess_filetype(filename, &ft) == GBM_ERR_OK)
        {
            // additionally check if the header can be decoded
            int     fd = -1;
            GBM     gbm;
            GBM_ERR gbmRc;

            if ( (fd = gbmAccessor.Gbm_io_open(filename, GBM_O_RDONLY)) == -1 )
            {
                return -1;
            }
            gbmRc = gbmAccessor.Gbm_read_header(filename, fd, ft, &gbm, "");

            gbmAccessor.Gbm_io_close(fd);

            if (gbmRc == GBM_ERR_OK)
            {
                return ft;
            }
        }

        // last chance is to test load the header
        int fd = -1;

        if ( (fd = gbmAccessor.Gbm_io_open(filename, GBM_O_RDONLY)) == -1 )
        {
            return -1;
        }

        int numFileTypes = -1;

        if (gbmAccessor.Gbm_query_n_filetypes(&numFileTypes) == GBM_ERR_OK)
        {
          GBM gbm;
          int formatCounter = 0;
          int foundFormat   = -1;

          for (ft = 0; ft < numFileTypes; ft++)
          {
            if (gbmAccessor.Gbm_read_header(filename, fd, ft, &gbm, "") == GBM_ERR_OK)
            {
                // There might be some formats that overlap, so count the
                // detected format types. If we find more than 1, we have
                // to decode the data to be sure to find the right format.
                formatCounter++;
                foundFormat = ft;
            }
          }

          gbmAccessor.Gbm_io_close(fd);

          if (formatCounter == 0)
          {
              // nothing found
              return -1;
          }

          if (formatCounter == 1)
          {
              // found the format, don't need to decode the data
              return foundFormat;
          }

          for (ft = 0; ft < numFileTypes; ft++)
          {
              // Now try to decode the data and take the first successfully decoded format.
              try
              {
                  GbmDocumentPage testPage(gbmAccessor);

                  // test loading always page 0
                  testPage.loadFromFile(filename, 0, ft, FALSE, FALSE);

                  // no exception -> we found the correct format
                  return ft;
              }
              catch(GbmException & ex)
              {
              }
          }
        }
    }

    return -1;
}

// ---------------------------------------------------------
// ---------------------------------------------------------

GbmDocument::GbmDocument(const GbmAccessor & accessor)
    : mGbmAccessor(accessor),
      renderers(NULL),
      pages(NULL),
      numPages(0),
      filename(NULL),
      force24bpp(FALSE),
      gbmFileType(-1),
      mRenderQueue(this)
{
    memset(&formatDescription, 0, sizeof(GBMFT));

    mFocusRendererInfo.pRenderer  = NULL;
    mFocusRendererInfo.scalerType = GbmRenderer::ScalerType_SIMPLE;
    mFocusRendererInfo.scaleX     = 0.0;
    mFocusRendererInfo.scaleY     = 0.0;
}

// ---------------------------------------------------------

GbmDocument::~GbmDocument()
{
    deleteAllPages();

    delete [] this->filename;
    this->filename = NULL;

    delete [] formatDescription.short_name;
    formatDescription.short_name = NULL;

    delete [] formatDescription.long_name;
    formatDescription.long_name = NULL;

    delete [] formatDescription.extensions;
    formatDescription.extensions = NULL;
}

// ---------------------------------------------------------

void GbmDocument::deleteAllPages()
{
    if (numPages > 0)
    {
        mRenderQueue.clear();

        mFocusRendererInfo.pRenderer  = NULL;
        mFocusRendererInfo.scalerType = GbmRenderer::ScalerType_SIMPLE;
        mFocusRendererInfo.scaleX     = 0.0;
        mFocusRendererInfo.scaleY     = 0.0;

        for (int p = 0; p < numPages; p++)
        {
            delete renderers[p];
            delete pages[p];
            renderers[p] = NULL;
            pages[p]     = NULL;
        }
        delete [] renderers;
        delete [] pages;
        renderers = NULL;
        pages     = NULL;
        numPages  = 0;
    }
}

// ---------------------------------------------------------

// Checks if the file contains a supported image format.
//
// If fileExtension is specified, only the filename is parsed
// the extension and then the extension is checked.
// If fileExtension is NULL, the file content is analyzed.

BOOL GbmDocument::isSupportedFormat(const char * filename,
                                    const char * fileExtension) const
{

    if (lookupGbmFileType(mGbmAccessor, filename, fileExtension) == -1)
    {
        return FALSE;
    }
    return TRUE;
}

// ---------------------------------------------------------

BOOL GbmDocument::hasContent() const
{
    return (numPages > 0);
}

// ---------------------------------------------------------

// Sets the file as active document. The existing document content is replaced.
//
// If fileExtension is specified, only the filename is parsed
// the extension and then the extension is checked.
// If fileExtension is NULL, the file content is analyzed.

void GbmDocument::setDocumentFile(const char * filename,
                                  const char * fileExtension,
                                  const BOOL   allowDelayedDataLoading,
                                  const BOOL   force24bpp) throw ( GbmException )
{
    delete [] this->filename;
    this->filename = NULL;

    // alloc memory for the filename and copy the filename to the buffer
    this->filename = new char[strlen(filename) + 1];
    strcpy(this->filename, filename);

    this->force24bpp = force24bpp;

    gbmFileType = lookupGbmFileType(mGbmAccessor, filename, fileExtension);
    if (gbmFileType == -1)
    {
        throw GbmException(GBM_ERR_NOT_SUPP, "File format is not supported");
    }

    deleteAllPages();

    // forward a thrown exception to the caller
    readFileFormatDescriptionFromFile();

    // forward a thrown exception to the caller
    const int filePages = readNumberOfPagesFromFile();

    // allocate the required number of pages
    try
    {
        renderers = new GbmRenderer*[filePages];
        pages     = new GbmDocumentPage*[filePages];
    }
    catch(...)
    {
    }
    // for compiler that don't throw std::bad_allow (e.g. IBM VAC)
    if ((renderers == NULL) || (pages == NULL))
    {
        deleteAllPages();
        throw GbmException(GBM_ERR_MEM);
    }

    numPages = filePages;

    // allocate page handler
    int allocatedPages = 0;
    try
    {
        for (int p = 0; p < filePages; p++)
        {
            pages[p] = NULL;
            pages[p] = new GbmDocumentPage(mGbmAccessor);
            if (pages[p] == NULL)
            {
                break;
            }
            allocatedPages++;
        }
    }
    catch(...)
    {
    }
    // for compiler that don't throw std::bad_allow (e.g. IBM VAC)
    if (allocatedPages != filePages)
    {
        deleteAllPages();
        throw GbmException(GBM_ERR_MEM);
    }

    // allocate page renderer
    allocatedPages = 0;
    try
    {
        for (int p = 0; p < numPages; p++)
        {
            renderers[p] = NULL;
            renderers[p] = new GbmRenderer();
            if (renderers[p] == NULL)
            {
                break;
            }
            allocatedPages++;
        }
    }
    catch(...)
    {
    }
    // for compiler that don't throw std::bad_allow (e.g. IBM VAC)
    if (allocatedPages != numPages)
    {
        deleteAllPages();
        throw GbmException(GBM_ERR_MEM);
    }

    if (allocatedPages > 0)
    {
        try
        {
            // read at least the first page to verify image data decoding
            pages[0]->loadFromFile(filename,
                                   0,
                                   gbmFileType,
                                   FALSE, // enforce full decoding
                                   force24bpp);

            // read the remaining pages which can use delayed data loading
            for (int p = 1; p < numPages; p++)
            {
                // forward a thrown exception to the caller
                pages[p]->loadFromFile(filename,
                                       p,
                                       gbmFileType,
                                       allowDelayedDataLoading,
                                       force24bpp);
            }
        }
        catch(GbmException & ex)
        {
            deleteAllPages();
            throw ex;
        }
    }

    if (numPages > 1)
    {
        if (mRenderQueue.getNumberOfBackgroundRenderPages() > 0)
        {
            mRenderQueue.start();
        }
    }
    else
    {
        mRenderQueue.stop();
    }
}

// ---------------------------------------------------------

void GbmDocument::writeDocumentPageToFile(const char * filename,
                                          const char * options,
                                          const int    pageNumber) throw ( GbmException )
{
    GbmDocumentPage & page(getPage(pageNumber));

    const BOOL needLock = (renderers[pageNumber] != mFocusRendererInfo.pRenderer);
    if (needLock)
    {
      mRenderQueue.lock();
    }
    try
    {
      page.writeToFile(filename, options);
    }
    catch(GbmException & ex)
    {
      if (needLock)
      {
        mRenderQueue.unlock();
      }
      throw ex;
    }
    if (needLock)
    {
      mRenderQueue.unlock();
    }
}

// ---------------------------------------------------------


void GbmDocument::getPageSize(const int pageNumber,
                                    int & pageWidth,
                                    int & pageHeight,
                                    int & pageColorDepth) throw ( GbmException )
{
    GbmDocumentPage & page(getPage(pageNumber));
    pageWidth      = page.getWidth();
    pageHeight     = page.getHeight();
    pageColorDepth = page.getColorDepth();
}

// ---------------------------------------------------------

void GbmDocument::rotate(const int pageNumber,
                         const int rotationAngle) throw ( GbmException )
{
    mRenderQueue.clear();

    GbmDocumentPage & page(getPage(pageNumber));
    page.rotate(rotationAngle);

    // reset all renderers
    for (int p = 0; p < numPages; p++)
    {
        renderers[p]->setRenderPage(NULL);
    }
    mFocusRendererInfo.pRenderer = NULL;
}

void GbmDocument::mirror(const int pageNumber,
                         const MirrorMode mode) throw ( GbmException )
{
    mRenderQueue.clear();

    GbmDocumentPage & page(getPage(pageNumber));
    page.mirror((GbmDocumentPage::MirrorMode) mode);

    // reset all renderers
    for (int p = 0; p < numPages; p++)
    {
        renderers[p]->setRenderPage(NULL);
    }
    mFocusRendererInfo.pRenderer = NULL;
}

// ---------------------------------------------------------

const GbmAccessor & GbmDocument::getGbmAccessor()
{
    return mGbmAccessor;
}

// ---------------------------------------------------------

void GbmDocument::setNumberOfBackgroundRenderPages(const int numRenderPages) throw ( GbmException )
{
    if (numRenderPages < 0)
    {
        throw GbmException(GBM_ERR_BAD_OPTION, "Number of background render pages must be > 0");
    }

    // reconfigure render queue (will clear all jobs)
    if (numRenderPages > 0)
    {
        if (! mRenderQueue.setNumberOfBackgroundRenderPages(numRenderPages + 1 /* the focus page added */))
        {
            throw GbmException(GBM_ERR_MEM, "Cannot allocate background renderers");
        }

        mRenderQueue.start();
    }
    else // == 0
    {
        mRenderQueue.stop();
    }
}

// ---------------------------------------------------------

void GbmDocument::resetRenderer(const int pageNumber) throw ( GbmException )
{
    if ((pageNumber < 0) || (pageNumber >= numPages))
    {
        throw GbmException(GBM_ERR_READ, "Requested page is not available");
    }

    const BOOL needLock = (renderers[pageNumber] != mFocusRendererInfo.pRenderer);
    if (needLock)
    {
      mRenderQueue.lock();
    }
    else
    {
      mFocusRendererInfo.pRenderer = NULL;
    }

    renderers[pageNumber]->setRenderPage(NULL);

    if (needLock)
    {
      mRenderQueue.unlock();
    }
}

// ---------------------------------------------------------

GbmRenderer & GbmDocument::getRenderer(const int pageNumber,
                                       const double scaleX, const double scaleY,
                                       const GbmRenderer::ScalerType scalerType) throw ( GbmException )
{
    if ((pageNumber < 0) || (pageNumber >= numPages))
    {
        throw GbmException(GBM_ERR_READ, "Requested page is not available");
    }

    // get the last focus renderer
    GbmRenderer * pLastFocusRenderer = mFocusRendererInfo.pRenderer;
    GbmRenderer * pNewFocusRenderer  = renderers[pageNumber];

    if (pLastFocusRenderer == pNewFocusRenderer)
    {
        // check if scale factors or scaler type has changed
        if ((scalerType == mFocusRendererInfo.scalerType) &&
            (scaleX     == mFocusRendererInfo.scaleX)     &&
            (scaleY     == mFocusRendererInfo.scaleY))
        {
           pNewFocusRenderer->setScalerType(scalerType);
           pNewFocusRenderer->prepareRenderCache(scaleX, scaleY);
           return *pNewFocusRenderer;
        }
    }

    // rebuild the background renderer queue

    mRenderQueue.clear();

    GbmDocumentPage & page(getPage(pageNumber));
    pNewFocusRenderer->setScalerType(scalerType);
    pNewFocusRenderer->setRenderPage(&page);
    pNewFocusRenderer->prepareRenderCache(scaleX, scaleY);

    mFocusRendererInfo.pRenderer  = pNewFocusRenderer;
    mFocusRendererInfo.scalerType = scalerType;
    mFocusRendererInfo.scaleX     = scaleX;
    mFocusRendererInfo.scaleY     = scaleY;

    // precalculate N pages backwards and forewards from requested page
    const int renderPages = (mRenderQueue.getNumberOfBackgroundRenderPages() > numPages)
                            ? numPages
                            : mRenderQueue.getNumberOfBackgroundRenderPages();
    if (renderPages > 0)
    {
        // schedule focus surrounding pages for background rendering
        int * surroundingPages = new int[renderPages];
        assert(surroundingPages != NULL);

        const int pagesAroundFocus = renderPages/2;
        int lastPage  = pageNumber + pagesAroundFocus;
        int firstPage = lastPage   - (renderPages - 1);

        if (firstPage < 0)
        {
            firstPage = 0;
            lastPage  = renderPages - 1;
        }
        if (lastPage >= numPages)
        {
            lastPage  = numPages - 1;
            firstPage = lastPage - (renderPages - 1);
            if (firstPage < 0) firstPage = 0;
        }

        assert(firstPage >= 0);
        assert(lastPage  < numPages);
        assert(lastPage - firstPage + 1 <= renderPages);

        int p;
        int numSurroundingPages = 0;

        for (int i = 1; i < renderPages; i++)
        {
            // prioritize the rendering of pages
            // - 1 after focus
            // - 1 before focus
            // - 2 after focus
            // - 2 before focus
            // ...

            p = pageNumber + i;
            if (p <= lastPage)
            {
                surroundingPages[numSurroundingPages] = p;
                numSurroundingPages++;
            }

            p = pageNumber - i;
            if (p >= firstPage)
            {
                surroundingPages[numSurroundingPages] = p;
                numSurroundingPages++;
            }
        }

        // free all rendered pages outside the prefetch range
        for (p = 0; p < firstPage; p++)
        {
            renderers[p]->setRenderPage(NULL);
        }
        for (p = lastPage + 1; p < numPages; p++)
        {
            renderers[p]->setRenderPage(NULL);
        }

        if (numSurroundingPages > 0)
        {
            BOOL success = mRenderQueue.schedulePages(surroundingPages,
                                                      numSurroundingPages,
                                                      scaleX, scaleY);
            assert(success);
            success = success; // suppress compiler warning
        }
        delete [] surroundingPages;
        surroundingPages = NULL;
    }
    else
    {
        // free previous render cache in case we have no background rendering
        if ((pLastFocusRenderer != NULL) &&
            (pLastFocusRenderer != mFocusRendererInfo.pRenderer))
        {
            pLastFocusRenderer->setRenderPage(NULL);
        }
    }

    // return the focus renderer
    return *(mFocusRendererInfo.pRenderer);
}

// ---------------------------------------------------------

// Forwarder for background renderer thread. Do not use otherwise!!!
void GbmDocument::prepareRenderCache(const int    pageNumber,
                                     const double scale_x,
                                     const double scale_y)
{
    try
    {
      GbmDocumentPage & page(getPage(pageNumber));

      GbmRenderer * pRenderer = renderers[pageNumber];
      pRenderer->setRenderPage(&page);
      pRenderer->setScalerType(mFocusRendererInfo.scalerType);
      pRenderer->prepareRenderCache(scale_x, scale_y);
    }
    catch(GbmException & ex)
    {
        // silently ignore
    }
}

// ---------------------------------------------------------

GbmDocumentPage & GbmDocument::getPage(const int pageNumber) throw ( GbmException )
{
    if ((pageNumber < 0) || (pageNumber >= numPages))
    {
        throw GbmException(GBM_ERR_READ, "Requested page is not available");
    }

    // check if the page has already been loaded
    if (! pages[pageNumber]->isValid())
    {
        mRenderQueue.lock();

        // unregister clear the renderer cache as we replace page data
        renderers[pageNumber]->setRenderPage(NULL);

        // load the page data

        // forward a thrown exception to the caller
        pages[pageNumber]->loadFromFile(filename, pageNumber, gbmFileType, TRUE, force24bpp);

        mRenderQueue.unlock();
    }

    return *pages[pageNumber];
}

// ---------------------------------------------------------

int GbmDocument::getNumberOfPages() const
{
    return numPages;
}

// ---------------------------------------------------------

const char * GbmDocument::getShortDescription() const
{
    if (hasContent())
    {
        return formatDescription.short_name;
    }
    return "";
}

const char * GbmDocument::getLongDescription() const
{
    if (hasContent())
    {
        return formatDescription.long_name;
    }
    return "";
}

const char * GbmDocument::getExtensions() const
{
    if (hasContent())
    {
        return formatDescription.extensions;
    }
    return "";
}

// ---------------------------------------------------------

void GbmDocument::readFileFormatDescriptionFromFile() throw ( GbmException )
{
    int     fd = -1;
    GBM_ERR gbmRc;

    delete [] formatDescription.short_name;
    formatDescription.short_name = NULL;

    delete [] formatDescription.long_name;
    formatDescription.long_name = NULL;

    delete [] formatDescription.extensions;
    formatDescription.extensions = NULL;

    GBMFT gbmft;
    gbmRc = mGbmAccessor.Gbm_query_filetype(gbmFileType, &gbmft);
    if (gbmRc != GBM_ERR_OK)
    {
        throw GbmException(gbmRc, "Can't extract file format description");
    }

    formatDescription.short_name = new char[strlen(gbmft.short_name) + 1];
    strcpy(formatDescription.short_name, gbmft.short_name);

    formatDescription.long_name = new char[strlen(gbmft.long_name) + 1];
    strcpy(formatDescription.long_name, gbmft.long_name);

    formatDescription.extensions = new char[strlen(gbmft.extensions) + 1];
    strcpy(formatDescription.extensions, gbmft.extensions);
}

// ---------------------------------------------------------

int GbmDocument::readNumberOfPagesFromFile() const throw ( GbmException )
{
    int     fd = -1;
    GBM_ERR gbmRc;

    if ( (fd = mGbmAccessor.Gbm_io_open(filename, GBM_O_RDONLY)) == -1 )
    {
        throw GbmException(GBM_ERR_READ, "Can't open file");
    }

    /* call gbm_read_imgcount(const char *fn, int fd, int ft, int *pimgcnt) */
    int pageCount = 1;

    gbmRc = mGbmAccessor.Gbm_read_imgcount(filename, fd, gbmFileType, &pageCount);
    if (gbmRc != GBM_ERR_OK)
    {
        mGbmAccessor.Gbm_io_close(fd);
        throw GbmException(gbmRc, "Can't read number of images");
    }
    mGbmAccessor.Gbm_io_close(fd);
    return pageCount;
}



